library tbwin;

uses
  Windows, Messages, mIRCStuff;

type
  TMouseEvents = (meMouse, meSingleClick, meDoubleClick, meUpClick, meLeave, meEnter);

  TMouseInfo = record
    X, Y: Integer;
    Down, Ctrl, Shift: Boolean;
    Hover: Boolean;
    Win: string[100];
  end;

  TWindowInfo = record
    WndProc: Pointer;
    Handle: HWND;
    MouseEvents: array[TMouseEvents] of TmString;
  end;

const
  // We use WM_MOUSELAST to know we aren't conflicting with mouse messages
  // that already exist.
  WM_TBMOUSELEAVE = WM_MOUSELAST + 1;
  WM_TBMOUSEENTER = WM_MOUSELAST + 2;

  // This is the identifier of the timer we use to detect mouse leave.
  MOUSE_TIMER = 999;

  MouseEvents: array[TMouseEvents] of Cardinal = (WM_MOUSEMOVE, WM_LBUTTONDOWN, WM_LBUTTONDBLCLK, WM_LBUTTONUP, WM_TBMOUSELEAVE, WM_TBMOUSEENTER);
  
  // @window names can only be up to 90 characters
  // long including the @.  We use 100 to be "safe."
  NAME_LEN = 100;
  CLASS_LEN = 20;

type
  TNameStr = array[0..NAME_LEN] of Char;
  TClassStr = array[0..CLASS_LEN] of Char;

// Returns the index of the window if the window is attached, otherwise
// returns -1.  Can use FindChild(handle) or FindChild('@name').
function FindChild(const Handle: HWND): Integer; overload; forward;
function FindChild(const Name: string): Integer; overload; forward;
function WinName(const Index: Integer): string; forward;

procedure AddChild(const Handle: HWND); forward;
procedure RemoveChild(const Index: Integer); forward;
procedure HandleMouseMessage(Msg: Cardinal; Index: Integer); forward;
procedure CleanUp; forward;
function MsgProc(hWnd: HWND; Msg: Cardinal; wParam: WPARAM; lParam: LPARAM): Integer; stdcall; forward;

var
  Wnds: array of TWindowInfo;

  CustWinClass: string;

  Hover: Integer = -1;
  
  // Let's go subclassin'
  ToolbarProc: Pointer = nil;
  ChildProc: Pointer = nil;
  WndProc: Pointer = nil;

  // Helper variable for use with tbtray.dll ...
  LoadedFirst: Boolean = False;

  // Handle to the MDI client area.  Needed for SetParent when the DLL is
  // unloaded...
  ClientHandle: HWND;

  // Handle to the toolbar.  Needed to subclass the toolbar.
  ToolbarHandle: HWND;

  // Information about where the mouse is.
  MouseInfo: TMouseInfo;

  ActiveWindow: Integer;

  // Event handlers.
  OnToolbarSize: string = '';

  FindName: string;

  {
    This is used to fix a bug in mIRC when using $dll() with relative and
    absolute paths to the same dll.  mIRC will think they are different files
    and try to load them twice.  Since Windows doesn't allow dlls to be loaded
    multiple times, this doesn't really cause a problem internally, but it
    does cause mIRC to fire LoadDll() multiple times and it can cause the dll's
    reference count to be incorrect.  Basically, you tell mIRC to keep on the
    first one but not any after that.
  }
  NeedUnload: Boolean;

// Basic stuff..
procedure LoadDll(var Info: LoadInfo); stdcall;
begin
  if (MainHandle = 0) and (Wnds = nil) then
  begin
    // KEEP this one
    Info.mKeep := True;
    SetLength(Wnds, 0);
    MainHandle := Info.mHwnd;
    ClientHandle := FindWindowEx(MainHandle, 0, 'MDIClient', nil);
  end
  else
  begin
    // mIRC bug fix...
    NeedUnload := True;
    // LOSE this one
    Info.mKeep := False;
  end;
end;

// Basic stuff..
function UnloadDll(mTimeout: Integer): Integer; stdcall;
begin
  Result := 0;

  // FIX THE BUG
  // When Info.mKeep is False, mIRC will immediately call UnloadDll.  If this
  // is the case, then we need to just exit.  Even though UnloadDll is being
  // called multiple times, the dll is still loaded... this bug fix just makes
  // sure that mIRC's internal reference count is correct.
  if (NeedUnload) then
  begin
    NeedUnload := False;
    Exit;
  end;
  
  if (mTimeout = 0) then
    CleanUp;
end;

function WinName(const Index: Integer): string;
var
  Buf: TNameStr;
begin
  Result := '';

  if (Index < 0) or (Index > Length(Wnds) - 1) then
    Exit;

  GetWindowText(Wnds[Index].Handle, Buf, NAME_LEN);

  Result := trim(Buf);

  if (Pos(' ', Result) > 0) then
    SetLength(Result, Pos(' ', Result) - 1);
  if (Result[1] <> '@') then
    Result := '@' + Result;
end;

procedure CleanUp;
var
  Res: Integer;
begin
  if (ToolbarHandle = 0) then
    Exit;

  while (Length(Wnds) > 0) do
    RemoveChild(Length(Wnds) - 1);

  // This is the code that makes sure tbwin.dll and tbtray.dll work together.
  Res := 0;
  if (LoadedFirst) then
    Res := SendMessage(ToolbarHandle, WM_NOTIFYTBWIN, Integer(WndProc), 1);

  if (Res <> 0) then
    RestoreMessageProc(ToolbarHandle, Pointer(Res))
  else
    RestoreMessageProc(ToolbarHandle, ToolbarProc);

  ToolbarHandle := 0;
  Hover := -1;
end;

function FindChild(const Handle: HWND): Integer; overload;
var
  I: Integer;
begin
  for I := 0 to Length(Wnds) - 1 do
  begin
    Result := I;
    if (Wnds[I].Handle = Handle) then
      Exit;
  end;
  Result := -1;
end;

function EnumChildProc(hWnd: HWND; lParam: LPARAM): BOOL; stdcall;
var
  BufA: TNameStr;
  BufB: TClassStr;
  Buf: string;
begin
  Result := True;

  GetWindowText(hWnd, BufA, NAME_LEN);
  GetClassName(hWnd, BufB, CLASS_LEN);

  Buf := trim(BufA);

  if (BufB = CustWinClass) then
  begin
    if (Pos(' ', Buf) > 0) then
      Buf := Copy(Buf, 1, Pos(' ', Buf) - 1);

    if (Buf[1] <> '@') then
      Buf := '@' + Buf;

    if (upper(Buf) = FindName) then
    begin
      Integer(Pointer(lParam)^) := hWnd;
      Result := False;
    end;
  end;
end;

function FindHandle(const Title: string): HWND;
begin
  Result := 0;
  FindName := upper(Title);
  EnumChildWindows(ClientHandle, @EnumChildProc, Longint(@Result));
end;

function FindChild(const Name: string): Integer; overload;
var
  BufA: TNameStr;
  I: Integer;
  Buf: string;
begin
  for I := 0 to Length(Wnds) - 1 do
  begin
    Result := I;

    GetWindowText(Wnds[I].Handle, BufA, NAME_LEN);

    Buf := trim(BufA);

    // Add a @ if they used -k to create the window.
    if (Buf[1] <> '@') then
      Buf := '@' + Buf;

    // Remove everything after the space, if there's a space.  After this
    // step, we should have a @name that matches the name used in mIRC.
    if (Pos(' ', Buf) > 0) then
      Delete(Buf, Pos(' ', Buf), MaxInt);

    if (upper(Buf) = upper(Name)) then
      Exit;
  end;
  Result := -1;
end;

procedure AddChild(const Handle: HWND);
var
  ChildNum: Integer;
begin
  ChildNum := Length(Wnds);
  SetLength(Wnds, ChildNum + 1);

  ActiveWindow := ChildNum;

  Wnds[ChildNum].Handle := Handle;
  Wnds[ChildNum].WndProc := NewMessageProc(Handle, @MsgProc);

  SetParent(Handle, ToolbarHandle);
  EnableWindow(Handle, False);
end;

procedure RemoveChild(const Index: Integer);
var
  I: Integer;
begin
  if (Index < 0) or (Index > (Length(Wnds) - 1)) then
    Exit;

  if (IsWindow(Wnds[Index].Handle)) then
  begin
    if (Wnds[Index].WndProc <> nil) then
      RestoreMessageProc(Wnds[Index].Handle, Wnds[Index].WndProc);

    SetParent(Wnds[Index].Handle, ClientHandle);
    EnableWindow(Wnds[Index].Handle, True);
  end;

  for I := Index to Length(Wnds) - 2 do
    Wnds[I] := Wnds[I + 1];

  SetLength(Wnds, Length(Wnds) - 1);

  ActiveWindow := 0;
end;

procedure HandleMouseMessage(Msg: Cardinal; Index: Integer);
var
  A: TMouseEvents;
begin
  // Iterate through the mouse events until we find the one we're looking for.
  // Using Low() and High() makes it easy to add or remove mouse events without
  // having to worry about the upper boundary.  MouseEvents is an array with the
  // elements of the TMouseEvents set.  Thus, we can reference them as 'meLeave'
  // or 'meSingleClick' instead of 0, 1, etc.
  for A := Low(TMouseEvents) to High(TMouseEvents) do

    // We found it.
    if (Msg = MouseEvents[A]) then
    begin

      // This is just some tidy-up code, it will help make mIRC's toolbar work
      // with our stuff a little better.
      if (Msg = WM_TBMOUSEENTER) then
        PostMessage(ToolbarHandle, WM_MOUSEMOVE, 0, makelong(500, 500));

      // If there's anything to do, well, do it.
      if (Wnds[Index].MouseEvents[A] <> '') then
        ExecCommand(Wnds[Index].MouseEvents[A]);

      // We found it, so we can get out.
      Exit;
    end;
end;

function MsgProc(hWnd: HWND; Msg: Cardinal; wParam: WPARAM; lParam: LPARAM): Integer; stdcall;
var
  P, X: TPoint;
  R: TRect;
  Index, I: Integer;
  TMO: TMouseInfo;
begin
  // Pretend we're gonna handle everything.
  Result := 0;

  if (hWnd <> ToolbarHandle) then
  begin
    Index := FindChild(hWnd);

    // This should never happen.
    if (Index = -1) then
    begin
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);
      Exit;
    end;

    if (Msg = WM_DESTROY) then
    begin
      RemoveChild(Index);

      if (Length(Wnds) = 0) then
        CleanUp;

      Result := DefWindowProc(hWnd, Msg, wParam, lParam);
    end

    else if (Msg = WM_TIMER) and (wParam = MOUSE_TIMER) then
    begin
      if (Hover = -1) then
      begin
        KillTimer(Wnds[Index].Handle, MOUSE_TIMER);
        Exit;
      end;

      GetCursorPos(P);
      GetWindowRect(Wnds[Index].Handle, R);

      if not (PtInRect(R, P)) then
      begin
        ScreenToClient(Wnds[Index].Handle, P);

        MouseInfo.X := P.X;
        MouseInfo.Y := P.Y;
        MouseInfo.Hover := False;
        MouseInfo.Win := WinName(Index);

        HandleMouseMessage(WM_TBMOUSELEAVE, Hover);
        Hover := -1;

        KillTimer(Wnds[Index].Handle, MOUSE_TIMER);
      end;

    end

    else
      Result := CallWindowProc(Wnds[Index].WndProc, hWnd, Msg, wParam, lParam);

    Exit;
  end;

  // tbtray.dll has just been loaded or unloaded.  That means the window
  // procedure for the toolbar has changed.  We need to update our saved one, so
  // if we unload, stuff doesn't mess up.
  if (Msg = WM_NOTIFYTBTRAY) then
  begin
    if not (LoadedFirst) and (wParam <> 0) then
    begin
      ToolbarProc := Pointer(wParam);
      Result := Integer(@MsgProc);
    end
    else
      Result := Integer(WndProc);

    if (lParam = 1) then
      LoadedFirst := True;
    Exit;
  end;

  case Msg of
    WM_LBUTTONDOWN,
    WM_LBUTTONUP,
    WM_LBUTTONDBLCLK,
    WM_MOUSEMOVE,
    WM_RBUTTONDOWN, WM_RBUTTONUP:
      begin
        P.X := LOWORD(lParam);
        P.Y := HIWORD(lParam);

        ClientToScreen(ToolbarHandle, P);

        MouseInfo.Down := Boolean(wParam and MK_LBUTTON);
        MouseInfo.Ctrl := Boolean(wParam and MK_CONTROL);
        MouseInfo.Shift := Boolean(wParam and MK_SHIFT);

        for I := Length(Wnds) - 1 downto 0 do
        begin
          GetWindowRect(Wnds[I].Handle, R);
          X := P;
          ScreenToClient(Wnds[I].Handle, X);

          // If the mouse is in the rectangle, we handle it.
          if (PtInRect(R, P)) then
          begin
            MouseInfo.X := X.X;
            MouseInfo.Y := X.Y;
            MouseInfo.Hover := True;
            MouseInfo.Win := WinName(I);

            if (Hover <> I) then
            begin
              SetTimer(Wnds[I].Handle, MOUSE_TIMER, 1, nil);

              if (Hover <> -1) then
              begin
                TMO := MouseInfo;
                MouseInfo.X := -1;
                MouseInfo.Y := -1;
                MouseInfo.Win := WinName(Hover);
                MouseInfo.Hover := False;
                KillTimer(Wnds[Hover].Handle, MOUSE_TIMER);
                HandleMouseMessage(WM_TBMOUSELEAVE, Hover);
                MouseInfo := TMO;
              end;

              Hover := I;
              HandleMouseMessage(WM_TBMOUSEENTER, Hover);
            end;

            LongRec(lParam).Lo := X.X;
            LongRec(lParam).Hi := X.Y;

            if (Msg = WM_RBUTTONDOWN) or (Msg = WM_RBUTTONUP) then
              SendMessage(Wnds[I].Handle, Msg, wParam, lParam)
            else
              HandleMouseMessage(Msg, I);

            Exit;
          end;
        end;
      end;
  end;

  Result := CallWindowProc(ToolbarProc, hWnd, Msg, wParam, lParam);

  // Check to see if this is a size message, and if the person has defined a
  // hook for it, then execute the hook, passing the width and height of the
  // toolbar as $1 and $2.
  if (Msg = WM_SIZE) and (OnToolbarSize <> '') and (OnToolbarSize[1] = '/') then
  begin
    GetClientRect(ToolbarHandle, R);
    ExecCommand(gettok(OnToolbarSize, ' ', 1) + ' ' + intstr(R.Right) + ' ' + intstr(R.Bottom));
  end;
end;

// OnMouse <event> [@window] [/command]
function OnMouse(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  S, W, M, Command: string;
  C: TmString;
  Index: Integer;
begin
  Result := 3;
  S := string(data);

  // Figure out what event we are hooking.
  W := gettok(S, ' ', 1);

  if (W[1] = '@') then
  begin
    M := gettok(S, ' ', 2);
    Command := gettok(S, ' ', -3);
    Index := FindChild(W);
    if (Index = -1) then
    begin
      Return(data, 'E_INVALID_WINDOW');
      Exit;
    end;
  end
  else
  begin
    M := W;
    Command := gettok(S, ' ', -2);
    Index := ActiveWindow;
  end;

  C := mstr(Command);

  if (M = 'mouse') then
    Wnds[Index].MouseEvents[meMouse] := C
  else if (M = 'sclick') then
    Wnds[Index].MouseEvents[meSingleClick] := C
  else if (M = 'dclick') then
    Wnds[Index].MouseEvents[meDoubleClick] := C
  else if (M = 'uclick') then
    Wnds[Index].MouseEvents[meUpClick] := C
  else if (M = 'leave') then
    Wnds[Index].MouseEvents[meLeave] := C
  else if (M = 'enter') then
    Wnds[Index].MouseEvents[meEnter] := C;

  Return(data, S_OK);
end;

function GetTBInfo(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  R: TRect;
  Handle: HWND;
begin
  Result := 3;

  Handle := FindWindowEx(mWnd, 0, 'mIRC_Toolbar', '');

  if not (IsWindow(Handle)) then
    Handle := FindWindowEx(mWnd, 0, 'mIRCTB32', '');

  if (IsWindow(Handle)) then
  begin
    GetClientRect(Handle, R);
    Return(data, intstr(R.Right) + ' ' + intstr(R.Bottom));
  end
  else
    Return(data, '-1 -1');
end;

function GetMouse(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  Key: Integer;
begin
  Result := 3;

  // Clear out the flags first.
  Key := 0;

  // These simulate $mouse.key by using bitwise OR to add the flags.  The Key
  // parameter can be used exactly like $mouse.key.
  if (MouseInfo.Down) then
    Key := Key or 1;
  if (MouseInfo.Ctrl) then
    Key := Key or 2;
  if (MouseInfo.Shift) then
    Key := Key or 4;

  // Returns <X Y Key Hover Win>
  Return(data,
   intstr(MouseInfo.X) + ' ' +
   intstr(MouseInfo.Y) + ' ' +
   intstr(Key) + ' ' +
   intstr(Ord(MouseInfo.Hover)) + ' ' +
   MouseInfo.Win
  );
end;

function OnSize(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  S: string;
begin
  S := string(data);

  // If it starts with a /, then he is setting a trigger.
  if (Length(S) > 0) and (S[1] = '/') then
    OnToolbarSize := S
  // Otherwise, remove the trigger.
  else
    OnToolbarSize := '';

  Result := 3;
  Return(data, S_OK);
end;

// Simple.
function Detach(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  Index: Integer;
begin
  Result := 3;

  Index := FindChild(string(data));

  if (Index = -1) then
  begin
    Return(data, 'E_INVALID_WINDOW');
    Exit;
  end;

  RemoveChild(Index);

  if (Length(Wnds) = 0) then
    CleanUp;

  Return(data, S_OK);
end;

function Select(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  S: string;
  Index: Integer;
begin
  Result := 3;
  S := string(data);

  Return(data, S_OK);

  Index := FindChild(S);
  if (Index <> -1) then
    ActiveWindow := Index
  else
    Return(data, 'E_INVALID_WINDOW');
end;

function Attach(mWnd, aWnd: HWND; data, params: PChar; a, b: BOOL): Integer; stdcall;
var
  Buf: TClassStr;
  S: string;
  E, C: Integer;
  Res: Integer;
begin
  Result := 3;
  S := string(data);

  if not (IsWindow(ToolbarHandle)) then
  begin
    ToolbarHandle := FindWindowEx(mWnd, 0, 'mIRC_Toolbar', '');
    CustWinClass := 'mIRC_Custom';

    if not (IsWindow(ToolbarHandle)) then
    begin
      ToolbarHandle := FindWindowEx(mWnd, 0, 'mIRCTB32', '');
      CustWinClass := 'custom';
    end;

    if not (IsWindow(ToolbarHandle)) then
    begin
      Return(data, 'E_CANNOT_LOCATE_TOOLBAR');
      Exit;
    end;


    Res := SendMessage(ToolbarHandle, WM_NOTIFYTBWIN, 0, 0);
    ToolbarProc := NewMessageProc(ToolbarHandle, @MsgProc);
    LoadedFirst := (Res = 0);
    if (LoadedFirst) then
      WndProc := ToolbarProc
    else
      WndProc := Pointer(Res);

    SetClassLong(ToolbarHandle, GCL_STYLE, GetClassLong(ToolbarHandle, GCL_STYLE) or CS_DBLCLKS);
  end;

  if (S[1] = '@') then
    C := FindHandle(S)
  else
  begin
    Val(S, C, E);
    if (E <> 0) then
      C := 0;
  end;

  if (IsWindow(C)) then
  begin
    GetClassName(C, Buf, CLASS_LEN);

    if (Buf = CustWinClass) and (ClientHandle <> 0) and (GetParent(C) = ClientHandle) then
    begin
      AddChild(C);
      Return(data, S_OK);
      Exit;
    end;
  end;

  Return(data, 'E_INVALID_WINDOW');
end;

exports
  LoadDll, UnloadDll,
  Attach, Select, Detach,
  OnMouse, OnSize,
  GetTBInfo, GetMouse;

begin
  MainHandle := 0;
  NeedUnload := False;
end.
